@@ -29,6 +29,7 @@
 (require 'seq)
 (require 'kv nil t)                     ;; May be absent.
 (require 'emacsql nil t)                ;; May be absent.
+(require 'emacsql-sqlite nil t)         ;; May be absent.
 
 ;;; Code:
 
@@ -154,20 +155,80 @@ easily debug into it.")
                                                                    ('builtin nil)
                                                                    ('emacsql '(:t t))))))))))
 
-(triples-deftest triples-test-select-predicate-object-fragment ()
-  (triples-test-with-temp-db
-    (triples-db-insert db 'sub1 'pred/foo "a whole phrase")
-    (should (equal
-             (mapcar (lambda (row) (seq-take row 3))
-                     (triples-db-select-predicate-object-fragment db 'pred/foo "whole"))
-             '((sub1 pred/foo "a whole phrase"))))))
-
 (triples-deftest triples-test-subjects-with-predicate-object ()
   (triples-test-with-temp-db
     (triples-db-insert db 'sub1 'pred/foo "bar")
     (should (equal (triples-subjects-with-predicate-object db 'pred/foo "bar")
                    '(sub1)))))
 
+(triples-deftest triples-test-db-select-pred-op-strings ()
+  (triples-test-with-temp-db
+    (triples-add-schema db 'named '(name))
+    (triples-set-subject db 123 '(named :name ("Foo" "Foo")))
+    (should (= 2 (length (triples-db-select-pred-op db :named/name '= "Foo"))))
+    (should (= 2 (length (triples-db-select-pred-op db :named/name 'like "F%"))))
+    (should (= 1 (length (triples-db-select-pred-op db :named/name '= "Foo" '(:index 0)))))
+    (should (= 2 (length (triples-db-select-pred-op db :named/name '> "A"))))
+    (should (= 0 (length (triples-db-select-pred-op db :named/name '< "A"))))
+    (should (= 2 (length (triples-db-select-pred-op db :named/name '= "foo"))))))
+
+(triples-deftest triples-test-db-select-pred-op-int ()
+  (triples-test-with-temp-db
+    (triples-add-schema db 'person '(age :base/unique t :base/type integer))
+    (triples-set-subject db 123 '(person :age 20))
+    (triples-set-subject db 456 '(person :age 40))
+    (triples-set-subject db 789 '(person :age 60))
+    (should (= 2 (length (triples-db-select-pred-op db :person/age '> 20))))
+    (should (= 3 (length (triples-db-select-pred-op db :person/age '>= 20))))
+    (should (= 2 (length (triples-db-select-pred-op db :person/age '!= 20))))
+    (should (= 3 (length (triples-db-select-pred-op db :person/age '!= 30))))
+    (should (= 1 (length (triples-db-select-pred-op db :person/age '= 20))))
+    (should (= 2 (length (triples-db-select-pred-op db :person/age '< 60))))
+    (should (= 3 (length (triples-db-select-pred-op db :person/age '<= 60))))
+    (should (= 0 (length (triples-db-select-pred-op db :person/age '> 60))))
+    ;; This may seem duplicative, but if we are doing textwise comparison, then
+    ;; this will fail.
+    (should (= 0 (length (triples-db-select-pred-op db :person/age '> 1000))))))
+
+(ert-deftest triples-test-builtin-emacsql-compat ()
+  (cl-loop for subject in '(1 a "a") do
+           (let ((triples-sqlite-interface 'builtin))
+             (triples-test-with-temp-db
+               (triples-add-schema db 'person
+                                     '(name :base/unique t :base/type string)
+                                     '(age :base/unique t :base/type integer))
+                 (triples-set-type db subject 'person :name "Alice Aardvark" :age 41)
+                 (should (equal (triples-get-type db subject 'person)
+                                '(:age 41 :name "Alice Aardvark")))
+                 (triples-close db)
+                 (let* ((triples-sqlite-interface 'emacsql)
+                        (db (triples-connect db-file)))
+                   (should (equal (triples-get-type db subject 'person)
+                                  '(:age 41 :name "Alice Aardvark")))
+                   (triples-close db))
+                 ;; Just so the last close will work.
+                 (setq db (triples-connect db-file))))))
+
+(ert-deftest triples-test-emacsql-builtin-compat ()
+  (cl-loop for subject in '(1 a "a") do
+           (let ((triples-sqlite-interface 'emacsql))
+             (triples-test-with-temp-db
+               (triples-add-schema db 'person
+                                     '(name :base/unique t :base/type string)
+                                     '(age :base/unique t :base/type integer))
+                 (triples-set-type db subject 'person :name "Alice Aardvark" :age 41)
+                 (should (equal (triples-get-type db subject 'person)
+                                '(:age 41 :name "Alice Aardvark")))
+                 (triples-close db)
+                 (let* ((triples-sqlite-interface 'builtin)
+                        (db (triples-connect db-file)))
+                   (should (equal (triples-get-type db subject 'person)
+                                  '(:age 41 :name "Alice Aardvark")))
+                   (triples-close db))
+                 ;; Just so the last close will work.
+                 (setq db (triples-connect db-file))))))
+    
+
 ;; After this we don't bother testing both with emacsql and the builtin sqlite,
 ;; since if the functions tested above work, it should also work for both.
 
@@ -183,7 +244,7 @@ easily debug into it.")
             '(locale :base/unique t)
             'alternate-names
             '(nicknames :base/unique nil))
-           '(replace-subject
+           '(replace-subject-type
              .
              ((named base/type schema)
               (named schema/property name)
@@ -220,7 +281,9 @@ easily debug into it.")
   (should (triples-test-op-equals
            (triples--set-type-op "Bert" 'named
                                  '(:name "Bertholomew The Second"
-                                         :alias ("Bert" "Berty")))
+                                         :alias ("Bert" "Berty"))
+                                 '((name :base/type string :base/unique t)
+                                   (alias :base/type string :base/unique nil)))
            '(replace-subject-type
              .
              (("Bert" base/type named)
@@ -229,15 +292,18 @@ easily debug into it.")
               ("Bert" named/alias "Berty" (:index 1)))))))
 
 (ert-deftest triples-schema-compliant ()
-  (triples-test-with-temp-db
-    (triples-add-schema db 'named
-                        '(name :base/unique t :base/type string)
-                        'alternate-names)
-    (triples-set-type db "foo" 'named :name "name")
-    (should (triples-verify-schema-compliant db '(("foo" named/name "bar"))))
-    (should-error (triples-verify-schema-compliant db '(("foo" named/name 5))))
-    (should-error (triples-verify-schema-compliant db '(("foo" named/name "bar" (:index 0)))))
-    (should (triples-verify-schema-compliant db '(("foo" named/alternate-names "bar" (:index 0)))))))
+  (let ((pal '((named/name :base/type string :base/unique t)
+               (named/alternate-names :base/type string :base/unique nil)
+               ;; Alias doesn't specify base/unique or base/type, so anything is fine.
+               (named/alias))))
+    (should (triples-verify-schema-compliant '(("foo" named/name "bar")) pal))
+    (should-error (triples-verify-schema-compliant '(("foo" named/name 5)) pal))
+    (should-error (triples-verify-schema-compliant '(("foo" named/name "bar" (:index 0))) pal))
+    (should (triples-verify-schema-compliant '(("foo" named/alternate-names "bar" (:index 0))) pal))
+    (should-error (triples-verify-schema-compliant '(("foo" named/alternate-names "bar" nil)) pal))
+    (should (triples-verify-schema-compliant '(("foo" named/alias "bar" nil)) pal))
+    (should (triples-verify-schema-compliant '(("foo" named/alias 5 nil)) pal))
+    (should (triples-verify-schema-compliant '(("foo" named/alias 5 (:index 0))) pal))))
 
 (defun triples-test-plist-sort (plist)
   "Sort PLIST in a standard way, for comparison."
@@ -268,7 +334,7 @@ easily debug into it.")
   (triples-test-with-temp-db
     (triples-add-schema db 'named
                         '(name :base/unique t))
-    (triples-add-schema db 'positioned '(position :/base/unique t))
+    (triples-add-schema db 'positioned '(position :base/unique t))
     (should-not (triples-get-subject db "foo"))
     (triples-set-subject db "foo"
                          '(named :name "bar")
@@ -281,7 +347,7 @@ easily debug into it.")
 (ert-deftest triples-set-types ()
   (triples-test-with-temp-db
     (triples-add-schema db 'named
-                        '(name :/base/unique t)
+                        '(name :base/unique t)
                        'alias)
     (triples-add-schema db 'reachable 'phone)
     (triples-set-type db "foo" 'named :name "Name" :alias '("alias1" "alias2"))
@@ -316,6 +382,18 @@ easily debug into it.")
                   (triples-get-type db "foo" 'embedding)))
    (should-error (triples-set-type db "foo" 'embedding :embedding '(1 2 3)))))
 
+(ert-deftest triples-cons ()
+  (triples-test-with-temp-db
+   (triples-add-schema db 'data '(data :base/unique t :base/type cons))
+   (triples-set-type db "foo" 'data :data '(a (b c)))
+   (should (equal '(:data (a (b c)))
+                  (triples-get-type db "foo" 'data)))
+   (should (= 1 (length (triples-db-select db nil 'data/data))))
+   ;; Let's also make sure if we store it as a straight list triples doesn't get
+   ;; confused and try to store it as separate rows in the db.
+   (triples-set-type db "foo" 'data :data '(a b c))
+   (should (= 1 (length (triples-db-select db nil 'data/data))))))
+
 (ert-deftest triples-reversed ()
   (triples-test-with-temp-db
    (triples-add-schema db 'named
@@ -367,8 +445,8 @@ easily debug into it.")
 
 (ert-deftest triples-move-subject ()
   (triples-test-with-temp-db
-   (triples-add-schema db 'named '(name))
-   (triples-add-schema db 'friend '(id))
+   (triples-add-schema db 'named '(name :base/unique t))
+   (triples-add-schema db 'friend '(id :base/unique t))
    (triples-set-subject db 123 '(named :name "Ada Lovelace"))
    (triples-set-subject db 456 '(named :name "Michael Faraday")
                         '(friend :id 123))
@@ -386,33 +464,41 @@ easily debug into it.")
     (triples-set-subject db 123 '(named :name ("Foo" "Foo")))
     (should (= 1 (length (triples-subjects-with-predicate-object db 'named/name "Foo"))))))
 
+(ert-deftest triples-test-schema-and-data-with-same-subject ()
+  (triples-test-with-temp-db
+   (triples-add-schema db 'foo '(bar))
+   (triples-set-subject db 'foo '(foo :bar "baz"))
+   (should (equal "baz" (plist-get (triples-get-subject db 'foo) :foo/bar)))
+   (triples-add-schema db 'foo '(bar))
+   (should (equal "baz" (plist-get (triples-get-subject db 'foo) :foo/bar)))))
+
 (ert-deftest triples-readme ()
   (triples-test-with-temp-db
-   (triples-add-schema db 'person
+    (triples-add-schema db 'person
        '(name :base/unique t :base/type string)
        '(age :base/unique t :base/type integer))
-   (triples-add-schema db 'employee
-       '(id :base/unique t :base/type integer)
-       '(manager :base/unique t)
-       '(reportees :base/virtual-reversed employee/manager))
-   ;; Set up catherine and dennis
-  (triples-set-type db "catherine" 'employee :manager "alice")
-  (triples-set-type db "dennis" 'employee :manager "alice")
-  (triples-delete-subject db "alice")
-  (triples-set-type db "alice" 'person :name "Alice Aardvark" :age 41)
-  (triples-set-type db "alice" 'employee :id 1901 :manager "bob")
-  (should (equal (triples-test-plist-sort (triples-get-subject db "alice"))
-                 (triples-test-plist-sort '(:person/name "Alice Aardvark" :person/age 41
-                                                         :employee/id 1901
-                                                         :employee/manager "bob"
-                                                         :employee/reportees ("catherine" "dennis")))))
-  (triples-set-subject db "alice" '(person :name "Alice Aardvark" :age 41)
-                       '(employee :id 1901 :manager "bob"))
-  (should (equal (triples-test-plist-sort (triples-get-subject db "alice"))
-                 (triples-test-plist-sort '(:person/name "Alice Aardvark" :person/age 41
-                                                         :employee/id 1901
-                                                         :employee/manager "bob"
-                                                         :employee/reportees ("catherine" "dennis")))))))
+    (triples-add-schema db 'employee
+                        '(id :base/unique t :base/type integer)
+                        '(manager :base/unique t)
+                        '(reportees :base/virtual-reversed employee/manager))
+    ;; Set up catherine and dennis
+    (triples-set-type db "catherine" 'employee :manager "alice")
+    (triples-set-type db "dennis" 'employee :manager "alice")
+    (triples-delete-subject db "alice")
+    (triples-set-type db "alice" 'person :name "Alice Aardvark" :age 41)
+    (triples-set-type db "alice" 'employee :id 1901 :manager "bob")
+    (should (equal (triples-test-plist-sort (triples-get-subject db "alice"))
+                   (triples-test-plist-sort '(:person/name "Alice Aardvark" :person/age 41
+                                                           :employee/id 1901
+                                                           :employee/manager "bob"
+                                                           :employee/reportees ("catherine" "dennis")))))
+    (triples-set-subject db "alice" '(person :name "Alice Aardvark" :age 41)
+                         '(employee :id 1901 :manager "bob"))
+    (should (equal (triples-test-plist-sort (triples-get-subject db "alice"))
+                   (triples-test-plist-sort '(:person/name "Alice Aardvark" :person/age 41
+                                                           :employee/id 1901
+                                                           :employee/manager "bob"
+                                                           :employee/reportees ("catherine" "dennis")))))))
 
 
 (provide 'triples-test)
