@@ -1,12 +1,12 @@
 ;;; triples.el --- A flexible triple-based database for use in apps  -*- lexical-binding: t; -*-
 
-;; Copyright (c) 2022  Free Software Foundation, Inc.
+;; Copyright (c) 2022, 2023  Free Software Foundation, Inc.
 
 ;; Author: Andrew Hyatt <ahyatt@gmail.com>
 ;; Homepage: https://github.com/ahyatt/triples
 ;; Package-Requires: ((seq "2.0") (emacs "28.1"))
 ;; Keywords: triples, kg, data, sqlite
-;; Version: 0.2.6
+;; Version: 0.3.1
 ;; This program is free software; you can redistribute it and/or
 ;; modify it under the terms of the GNU General Public License as
 ;; published by the Free Software Foundation; either version 2 of the
@@ -65,11 +65,7 @@ If FILE is nil, use `triples-default-database-filename'."
   (let ((file (or file triples-default-database-filename)))
     (pcase triples-sqlite-interface
       ('builtin (let* ((db (sqlite-open file)))
-                  (sqlite-execute db "CREATE TABLE IF NOT EXISTS triples(subject TEXT NOT NULL, predicate TEXT NOT NULL, object NOT NULL, properties TEXT NOT NULL)")
-                  (sqlite-execute db "CREATE INDEX IF NOT EXISTS subject_idx ON triples (subject)")
-                  (sqlite-execute db "CREATE INDEX IF NOT EXISTS subject_predicate_idx ON triples (subject, predicate)")
-                  (sqlite-execute db "CREATE INDEX IF NOT EXISTS predicate_object_idx ON triples (predicate, object)")
-                  (sqlite-execute db "CREATE UNIQUE INDEX IF NOT EXISTS subject_predicate_object_properties_idx ON triples (subject, predicate, object, properties)")
+                  (triples-setup-table-for-builtin db)
                   db))
       ('emacsql
        (require 'emacsql)
@@ -89,6 +85,16 @@ If FILE is nil, use `triples-default-database-filename'."
            (emacsql db [:create-unique-index subject_predicate_object_properties_idx :on triples [subject predicate object properties]]))
          db)))))
 
+(defun triples-setup-table-for-builtin (db)
+  "Set up the triples table in DB.
+This is a separate function due to the need to use it during
+upgrades to version 0.3"
+  (sqlite-execute db "CREATE TABLE IF NOT EXISTS triples(subject NOT NULL, predicate TEXT NOT NULL, object NOT NULL, properties TEXT NOT NULL)")
+  (sqlite-execute db "CREATE INDEX IF NOT EXISTS subject_idx ON triples (subject)")
+  (sqlite-execute db "CREATE INDEX IF NOT EXISTS subject_predicate_idx ON triples (subject, predicate)")
+  (sqlite-execute db "CREATE INDEX IF NOT EXISTS predicate_object_idx ON triples (predicate, object)")
+  (sqlite-execute db "CREATE UNIQUE INDEX IF NOT EXISTS subject_predicate_object_properties_idx ON triples (subject, predicate, object, properties)"))
+
 (defun triples-close (db)
   "Close sqlite database DB."
   (pcase triples-sqlite-interface
@@ -137,15 +143,17 @@ exist at any time. Older backups are the ones that are deleted."
   "If VAL is a string, return it as enclosed in quotes
 This is done to have compatibility with the way emacsql stores
 values. Turn a symbol into a string as well, but not a quoted
-one, because sqlite cannot handle symbols."
+one, because sqlite cannot handle symbols. Integers do not need
+to be stringified."
   ;; Do not print control characters escaped - we want to get things out exactly
   ;; as we put them in.
   (let ((print-escape-control-characters nil))
-    (if val
-        (format "%S" val)
+    (pcase val
       ;; Just to save a bit of space, let's use "()" instead of "null", which is
       ;; what it would be turned into by the pcase above.
-      "()")))
+      ((pred null) "()")
+      ((pred integerp) val)
+      (_ (format "%S" val)))))
 
 (defun triples-standardize-result (result)
   "Return RESULT in standardized form.
@@ -242,6 +250,49 @@ all to nil, everything will be deleted, so be careful!"
     ('emacsql (emacsql db [:delete :from triples :where (= subject $s1) :and (like predicate $r2)]
                        subject (format "%s/%%" (triples--decolon pred-prefix))))))
 
+(defun triples-db-select-pred-op (db pred op val &optional properties)
+  "Select triples matching predicates with PRED having OP relation to VAL.
+OP is a comparison operator, and VAL is the value to compare. OP,
+the comparison operator, is a symbol for a standard numerical
+comparison such as `=', `!=', `>', or, when `val' is a strings,
+`like'.  All alphabetic comparison is case insensitive.
+
+If PROPERTIES is given, triples must match the given properties."
+  (unless (symbolp pred)
+    (error "Predicates in triples must always be symbols"))
+  (let ((pred (triples--decolon pred)))
+    (pcase triples-sqlite-interface
+      ('builtin
+       (mapcar (lambda (row) (mapcar #'triples-standardize-result row)) 
+               (sqlite-select
+                db
+                (concat "SELECT * FROM triples WHERE predicate = ? AND  "
+                        (if (numberp val)
+                            "CAST(object AS INTEGER) "
+                          "object COLLATE NOCASE ")
+                        (symbol-name op) " ?"
+                        (when properties " AND properties = ?"))
+                (append
+                 (list (triples-standardize-val pred)
+                       (triples-standardize-val val))
+                 (when properties (list (triples-standardize-val properties)))))))
+      ('emacsql
+       (emacsql db
+                (append
+                 [:select * :from triples :where (= predicate $s1) :and]
+                 (pcase op
+                   ('< [(< object $s2)])
+                   ('<= [(<= object $s2)])
+                   ('= [(= object $s2)])
+                   ('!= [(!= object $s2)])
+                   ('>= [(>= object $s2)])
+                   ('> [(> object $s2)])
+                   ('like [(like object $s2)]))
+                 (when (stringp val) [:collate :nocase])
+                 (when properties
+                   (list :and '(= properties $s3))))
+                pred val properties)))))
+
 (defun triples-db-select-pred-prefix (db subject pred-prefix)
   "Return rows matching SUBJECT and PRED-PREFIX."
   (pcase triples-sqlite-interface
@@ -252,16 +303,6 @@ all to nil, everything will be deleted, so be careful!"
     ('emacsql (emacsql db [:select * :from triples :where (= subject $s1) :and (like predicate $r2)]
                        subject (format "%s/%%" pred-prefix)))))
 
-(defun triples-db-select-predicate-object-fragment (db predicate object-fragment)
-  "Return rows with PREDICATE and with OBJECT-FRAGMENT in object."
-  (pcase triples-sqlite-interface
-    ('builtin (mapcar (lambda (row) (mapcar #'triples-standardize-result row))
-                      (sqlite-select db "SELECT * from triples WHERE predicate = ? AND object LIKE ?"
-                                     (list (triples-standardize-val predicate)
-                                           (format "%%%s%%" object-fragment)))))
-    ('emacsql (emacsql db [:select * :from triples :where (= predicate $s1) :and (like object $s2)]
-                       predicate (format "%%%s%%" object-fragment)))))
-
 (defun triples-db-select (db &optional subject predicate object properties selector)
   "Return rows matching SUBJECT, PREDICATE, OBJECT, PROPERTIES.
 If any of these are nil, they are not included in the select
@@ -379,12 +420,14 @@ merged into NEW-SUBJECT."
   (mapcar #'car
           (triples-db-select db type 'schema/property nil nil '(object))))
 
-(defun triples-verify-schema-compliant (db triples)
-  "Error if TRIPLES is not compliant with schema in DB."
+(defun triples-verify-schema-compliant (triples prop-schema-alist)
+  "Error if TRIPLES is not compliant with schema in PROP-SCHEMA-ALIST.
+PROP-SCHEMA-ALIST is an alist of the relevant properties to the
+data stored, in combined type/property form, and their schema
+definitions."
   (mapc (lambda (triple)
-          (pcase-let ((`(,type . ,prop) (triples-combined-to-type-and-prop (nth 1 triple))))
-            (unless (or (eq type 'base)
-                        (triples-db-select db type 'schema/property prop nil))
+          (pcase-let ((`(,type . ,_) (triples-combined-to-type-and-prop (nth 1 triple))))
+            (unless (or (eq type 'base) (assoc (nth 1 triple) prop-schema-alist))
               (error "Property %s not found in schema" (nth 1 triple)))))
         triples)
   (mapc (lambda (triple)
@@ -393,7 +436,7 @@ merged into NEW-SUBJECT."
                                                           (triples--decolon pred-prop)))))
                                  (if (fboundp f)
                                      (funcall f val triple))))
-                             (triples-properties-for-predicate db (nth 1 triple)))) triples))
+                               (cdr (assoc (nth 1 triple) prop-schema-alist)))) triples))
 
 (defun triples-add-schema (db type &rest props)
   "Add schema for TYPE and its PROPS to DB."
@@ -404,7 +447,7 @@ merged into NEW-SUBJECT."
 PROPS is a list of either property symbols, or lists of
 properties of the type and the meta-properties associated with
 them."
-  (cons 'replace-subject
+  (cons 'replace-subject-type
         (cons `(,type base/type schema)
               (cl-loop for p in props
                        nconc
@@ -424,8 +467,21 @@ them."
 (defun triples-set-type (db subject type &rest properties)
   "Create operation to replace PROPERTIES for TYPE for SUBJECT in DB.
 PROPERTIES is a plist of properties, without TYPE prefixes."
-  (let ((op (triples--set-type-op subject type properties)))
-    (triples-verify-schema-compliant db (cdr op))
+  (let* ((prop-schema-alist
+          (mapcar (lambda (prop)
+                    (cons (triples--decolon prop)
+                          (triples-properties-for-predicate
+                           db
+                           (triples-type-and-prop-to-combined type prop))))
+                  (triples--plist-mapcar (lambda (k _) k) properties)))
+         (op (triples--set-type-op subject type properties prop-schema-alist)))
+    (triples-verify-schema-compliant
+     (cdr op)
+     ;; triples-verify-schema-compliant can act on triples from many types, so
+     ;; we have to include the type information in our schema property alist.
+     (mapcar (lambda (c)
+               (cons (triples-type-and-prop-to-combined type (car c))
+                     (cdr c))) prop-schema-alist))
     (triples--add db op)))
 
 (defmacro triples-with-transaction (db &rest body)
@@ -473,21 +529,27 @@ given in the COMBINED-PROPS will be removed."
       (cl-loop for k being the hash-keys of type-to-plist using (hash-values v)
                do (apply #'triples-set-type db subject k v)))))
 
-(defun triples--set-type-op (subject type properties)
+(defun triples--set-type-op (subject type properties type-schema)
   "Create operation to replace PROPERTIES for TYPE for SUBJECT.
-PROPERTIES is a plist of properties, without TYPE prefixes."
+PROPERTIES is a plist of properties, without TYPE prefixes.
+TYPE-SCHEMA is an alist of property symbols to their schema,
+which is necessary to understand when lists are supposed to be
+broken down into separate rows, and when to leave as is."
   (cons 'replace-subject-type
         (cons (list subject 'base/type type)
               (triples--plist-mapcan
                (lambda (prop v)
-                 (if (listp v)
+                 (let ((prop-schema (cdr (assoc (triples--decolon prop) type-schema))))
+                   (if (and
+                        (listp v)
+                        (not (plist-get prop-schema :base/unique)))
                      (cl-loop for e in v for i from 0
                               collect
                               (list subject
                                     (triples-type-and-prop-to-combined type prop)
                                     e
                                     (list :index i)))
-                   (list (list subject (triples-type-and-prop-to-combined type prop) v))))
+                   (list (list subject (triples-type-and-prop-to-combined type prop) v)))))
                properties))))
 
 (defun triples-get-type (db subject type)
@@ -557,7 +619,7 @@ data you own with `triples-remove-type'."
 
 (defun triples-search (db cpred text)
   "Search DB for instances of combined property CPRED with TEXT."
-  (triples-db-select-predicate-object-fragment db cpred text))
+  (triples-db-select-pred-op db cpred 'like (format "%%%s%%" text)))
 
 (defun triples-with-predicate (db cpred)
   "Return all triples in DB with CPRED as its combined predicate."
